#include "Global.h"
#include "Util.h"
#include "Object.h"
#include "ObjectIlk.h"
#include "Creature.h"
#include "Hero.h"

static int loadRayGun (shObject *gun);

void
initializeRayGuns (void)
{
    int n = 0;

    /* enum: kObjGenericRayGun */
    /* enum: kObjFirstRayGun = kObjGenericRayGun */
    NEW_RAY_GUN(kObjGenericRayGun,
        "ray gun", "ray gun", RayGunData[n].mColor, 0,
        NULL, 200 BZ, ABSTRACT);
    AllIlks[kObjGenericRayGun]->mParent = kObjNothing;

    /* enum: kObjEmptyRayGun */
    NEW_RAY_GUN(kObjEmptyRayGun,
        "empty ray gun", RayGunData[n].mDesc, RayGunData[n].mColor, 0,
        NULL, 200 BZ, 10);
    AllIlks[kObjEmptyRayGun]->mUseFunc = loadRayGun;
    AllIlks[kObjEmptyRayGun]->mFlags &= kChargeable;
    n++;

    /* enum: kObjFreezeRayGun */
    NEW_RAY_GUN(kObjFreezeRayGun,
        "freeze ray gun", RayGunData[n].mDesc, RayGunData[n].mColor, 0,
        newRayGunAttack (shAttack::kFreezeRay, kFreezing, 4, 32),
        800 BZ, 10);
    n++;

    /* enum: kObjDisintegrationRayGun */
    NEW_RAY_GUN(kObjDisintegrationRayGun,
        "disintegration ray gun", RayGunData[n].mDesc, RayGunData[n].mColor, 0,
        newRayGunAttack (shAttack::kDisintegrationRay, kDisintegrating, 20, 40),
        2000 BZ, 1);
    n++;
    AllIlks[kObjDisintegrationRayGun]->mMissileAttack =
        newGrenadeAttack (shAttack::kBlast, 1, kDisintegrating, 20, 40);

    /* enum: kObjHeatRayGun */
    NEW_RAY_GUN(kObjHeatRayGun,
        "heat ray gun", RayGunData[n].mDesc, RayGunData[n].mColor, 0,
        newRayGunAttack (shAttack::kHeatRay, kBurning, 4, 32),
        800 BZ, 15);
    n++;

    /* enum: kObjGaussRayGun */
    NEW_RAY_GUN(kObjGaussRayGun,
        "gauss ray gun", RayGunData[n].mDesc, RayGunData[n].mColor, 0,
        newRayGunAttack (shAttack::kGaussRay, kMagnetic, 4, 32),
        800 BZ, 12);
    n++;

    /* enum: kObjPoisonRayGun */
    NEW_RAY_GUN(kObjPoisonRayGun,
        "poison ray gun", RayGunData[n].mDesc, RayGunData[n].mColor, 0,
        newRayGunAttack (shAttack::kPoisonRay, kPoisonous, 1, 4),
        800 BZ, 12);
    n++;

    /* enum: kObjGammaRayGun */
    NEW_RAY_GUN(kObjGammaRayGun,
        "gamma ray gun", RayGunData[n].mDesc, RayGunData[n].mColor,
        kRadioactive,
        newRayGunAttack (shAttack::kGammaRay, kRadiological, 4, 32),
        800 BZ, 10);
    n++;

    /* enum: kObjStasisRayGun */
    NEW_RAY_GUN(kObjStasisRayGun,
        "stasis ray gun", RayGunData[n].mDesc, RayGunData[n].mColor, 0,
        newRayGunAttack (shAttack::kStasisRay, kParalyzing, 1, 6),
        800 BZ, 5);
    n++;

    /* enum: kObjTransporterRayGun */
    NEW_RAY_GUN(kObjTransporterRayGun,
        "transporter ray gun", RayGunData[n].mDesc, RayGunData[n].mColor, 0,
        newRayGunAttack (shAttack::kTransporterRay, kTransporting, 4, 32),
        800 BZ, 15);
    n++;

    /* enum: kObjHealingRayGun */
    NEW_RAY_GUN(kObjHealingRayGun,
        "healing ray gun", RayGunData[n].mDesc, RayGunData[n].mColor, 0,
        newRayGunAttack (shAttack::kHealingRay, kHealing, 4, 32),
        800 BZ, 5);
    n++;

    /* enum: kObjRestorationRayGun */
    /* enum: kObjLastRayGun = kObjRestorationRayGun */
    NEW_RAY_GUN(kObjRestorationRayGun,
        "restoration ray gun", RayGunData[n].mDesc, RayGunData[n].mColor, 0,
        newRayGunAttack (shAttack::kRestorationRay, kRestoring, 4, 32),
        800 BZ, 5);
    n++;
}


char *
shObjectIlk::getRayGunColor ()
{
    static char buf[40];
    char *p;

    strcpy (buf, mAppearance);
    for (p = &buf[0]; ' ' != *p; ++p)
        ;
    *p = 0;
    return &buf[0];
}



/* returns time elapsed */

static int
loadRayGun (shObject *gun)
{
    shObjectVector v;
    shObject *can;

    selectObjects (&v, Hero.mInventory, kCanister);
    can = Hero.quickPickItem (&v, "reload with", 0);

    if (!can) {
        return 0;
    }
    if        (can->isA (kObjLNO)) {
        gun->mIlkId = kObjFreezeRayGun;
    } else if (can->isA (kObjNapalm)) {
        gun->mIlkId = kObjHeatRayGun;
    } else if (can->isA (kObjAntimatter)) {
        gun->mIlkId = kObjDisintegrationRayGun;
    } else if (can->isA (kObjPoisonCanister)) {
        gun->mIlkId = kObjPoisonRayGun;
    } else if (can->isA (kObjPlasmaCanister)) {
        gun->mIlkId = kObjGaussRayGun;
    } else if (can->isA (kObjMutagenCanister)) {
        gun->mIlkId = kObjGammaRayGun;
    } else if (can->isA (kObjSpiceMelange)) {
        gun->mIlkId = kObjTransporterRayGun;
    } else if (can->isA (kObjSuperGlue)) {
        gun->mIlkId = kObjStasisRayGun;
    } else if (can->isA (kObjHealingCanister) ||
               can->isA (kObjFullHealingCanister))
    {
        gun->mIlkId = kObjHealingRayGun;
    } else if (can->isA (kObjRestorationCanister)) {
        gun->mIlkId = kObjRestorationRayGun;
    } else {
        /* canisters that don't load ray guns right now:
           beer
           nano cola
           water
           rad-away
           universal solvent - damages player besides ruining ray gun?
           speed - acceleration ray?
           gain ability
           canned alien embryo
        */
        I->p ("Your ray gun is ruined!");
        can->maybeName ();
        Hero.useUpOneObjectFromInventory (can);
        Hero.useUpOneObjectFromInventory (gun);
        return FULLTURN;
    }
    gun->mCharges = NDX (2, 6);
    if (!Hero.isBlind ()) {
        I->p ("The light on the ray gun is %s now.", gun->myIlk ()->getRayGunColor ());
    } else {
        gun->resetAppearanceKnown ();
    }

    if (can->isIlkKnown ()) {
        gun->setIlkKnown ();
        gun->setAppearanceKnown ();
    } else if (gun->isIlkKnown () && !Hero.isBlind ()) {
        can->setIlkKnown ();
    } else {
        can->maybeName ();
    }

    Hero.useUpOneObjectFromInventory (can);
    gun->setAppearanceKnown ();

    return FULLTURN;
}
