#include <ctype.h>

#include "Global.h"
#include "Util.h"
#include "Map.h"
#include "AUI.h" // argh
#include "Hero.h"
#include "Mutant.h"
#include "Game.h"

#include <curses.h>
#ifndef _PANEL_H
extern "C" {
#include <panel.h>
}
#define _PANEL_H
#endif

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifdef _WIN32
#include <io.h>
#else
#include <unistd.h>
#endif
#include <fcntl.h>
#include <errno.h>
#include <time.h>

#ifndef O_BINARY
#define O_BINARY 0
#endif

extern int GameOver;
/* calculation of score:

1 point per XP earned
500 points per level visited besides level 1
10000 for finding the bizarro orgasmatron

*/

static int
copystr (char *dest, char *src, int len)
{
    int i;

    for (i = 0; i < len; i++) {
        dest[i] = src[i];
        if (0 == src[i]) {
            return -1;
        }
    }
    while (i > 0) {
        --i;
        if (isspace (src[i])) {
            dest[i] = 0;
            return i+1;
        }
    }
    dest[i] = 0;
    return len;
}

int
shHero::tallyScore ()
{
    return mScore;
}


const char *
shMapLevel::getDescription ()
{
    switch (Level->mType) {
    case shMapLevel::kTown: return "in Robot Town";
    case shMapLevel::kRadiationCave: return "in the Gamma Caves";
    case shMapLevel::kMainframe: return "in the Mainframe";
    case shMapLevel::kRabbit: return "in the Rabbit's Hole";
    case shMapLevel::kSewer: return "in the sewer";
    case shMapLevel::kSewerPlant:
        return "in the waste treatment plant";
    default:
        return "in the space base";
    }
}


void
shHero::epitaph (char *buf, int len, shCauseOfDeath how, const char *killer)
{
    switch (how) {
    case kSlain:
    case kKilled:
        if (killer && !strcasecmp (killer, "a dalek")) {
            snprintf (buf, len, "Exterminated by %s", killer);
        } else if (killer && !strcasecmp (killer, "a man eating plant")) {
            snprintf (buf, len, "Devoured by %s", killer);
        } else {
            snprintf (buf, len, "Killed by %s", killer ? killer : "a monster");
        }
        break;
    case kAnnihilated:
        snprintf (buf, len, "Annihilated by %s",
                  killer ? killer : "a monster");
        break;
    case kSuddenDecompression:
        snprintf (buf, len, "Died of vaccuum exposure");
        break;
    case kTransporterAccident:
        snprintf (buf, len, "Died in a transporter accident");
        break;
    case kSuicide:
        snprintf (buf, len, "Committed suicide");
        break;
    case kDrowned:
        snprintf (buf, len, "Drowned %s", killer);
        break;
    case kBrainJarred:
        snprintf (buf, len, "Brain-napped by %s", killer);
        break;
    case kMisc:
        snprintf (buf, len, "%s", killer);
        break;
    case kQuitGame:
        snprintf (buf, len, "Quit");
        break;
    case kWonGame:
        snprintf (buf, len, "Activated the Bizarro Orgasmatron");
        break;
    }

    snprintf (&buf[strlen (buf)], len, " %s at depth %d.",
              Level->getDescription (), Level->mDLevel);
}



void
shHero::tomb (char *message)
{
    char buf1[80];
    char buf2[80];
    int n, l;
    int row;
    WINDOW *win;
    PANEL *panel;

    win = newwin (25, 60, 0, 0);
    if (!win) {
        I->p ("Couldn't create tombstone.");
        return;
    }
    panel = new_panel (win);

    wattrset (win, ColorMap[kBrightCyan]);

    row = 3;

    mvwaddstr (win, row++, 6, "            _________            ");
    mvwaddstr (win, row++, 6, "           /         \\           ");
    mvwaddstr (win, row++, 6, "          /  _     _  \\          ");
    mvwaddstr (win, row++, 6, "         /  | \\ | | \\  \\         ");
    mvwaddstr (win, row++, 6, "        /   |_/ | |_/   \\        ");
    mvwaddstr (win, row++, 6, "       /    | \\ | |      \\       ");
    mvwaddstr (win, row++, 6, "      |                   |      ");
    mvwaddstr (win, row++, 6, "      |                   |      ");
    mvwaddstr (win, row++, 6, "      |                   |      ");
    mvwaddstr (win, row++, 6, "      |                   |      ");
    mvwaddstr (win, row++, 6, "      |                   |      ");
    mvwaddstr (win, row++, 6, "      |                   |      ");
    mvwaddstr (win, row++, 6, "      |                   |      ");
    mvwaddstr (win, row++, 6, "      |                   |      ");
    mvwaddstr (win, row++, 6, "      |                   |      ");
    mvwaddstr (win, row++, 6, "      |                   |      ");
    mvwaddstr (win, row++, 2,
               "----------+-------------------+--------------------------");

    row -= 10;

    wattrset (win, ColorMap[kWhite]);
    snprintf (buf1, 20, "%s", mName);
    mvwaddstr (win, row, 22 - strlen (buf1) / 2, buf1);

    snprintf (buf1, 18, "%s", getTitle ());
    wattrset (win, ColorMap[kGray]);
    mvwaddstr (win, ++row, 22 - strlen (buf1) / 2, buf1);

    row += 2;
    n = 0;
    do {
        l = copystr (buf2, &message[n], 18);
        n += l;
        mvwaddstr (win, row, 22 - strlen (buf2) / 2, buf2);
        row++;
    } while (l >= 0);

    wmove (win, 0, 0);

    update_panels ();
    doupdate();
    I->getChar ();
    hide_panel (panel);
    del_panel (panel);
    delwin (win);
    update_panels ();
    I->drawScreen ();
}


void
readHighScoreTable (HighScoreEntry scores[NUMHIGHSCORES], int *nscores, int *entry)
{
    char filename[PRIME_PATH_LENGTH];
    snprintf (filename, sizeof(filename)-1, "%s/highscores.dat", SCOREDIR);
    int fd = open (filename, O_RDONLY | O_BINARY,
#ifdef _WIN32
        S_IREAD | S_IWRITE);
#else
        S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
#endif

    *nscores = 0;
    *entry = 0;
    if (-1 != fd) {
        int numread;
        while (numread = read (fd, &scores[*nscores], sizeof (HighScoreEntry)),
               numread == sizeof (HighScoreEntry))
        {
            if (Hero.tallyScore () < scores[*nscores].mScore) {
                *entry = *nscores + 1;
            }
            ++(*nscores);
        }
        close (fd);
        if (numread == -1) {
            I->p ("The high score table is corrupt and will be made anew.");
            *nscores = 0;
            *entry = 0;
            unlink (filename);
        }
    }
    /* Else create a new high score table. */
}


void
shHero::logGame (char *message)
{
    char logname[PRIME_PATH_LENGTH];
    int fd;
    time_t clocktime;
    struct tm *when;

    time (&clocktime);
    when = localtime (&clocktime);

    snprintf (logname, sizeof(logname)-1, "%s/logfile.txt", SCOREDIR);

    fd = open (logname, O_WRONLY | O_APPEND | O_CREAT,
#ifdef _WIN32
               S_IREAD | S_IWRITE);
#else
               S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
#endif

    if (-1 == fd) {
        I->p ("Couldn't open logfile.");
        I->pause ();
    } else {
        char logstr[1024];
        snprintf (logstr, sizeof(logstr),
                  "%s %04d-%02d-%02d %10d %2d %6d %s the %s %s%s\n",
                  PRIME_VERSION,
                  when->tm_year + 1900, when->tm_mon + 1, when->tm_mday,
                  mScore, mCLevel, mXP, mName,
                  getTitle (),
                  GodMode? "(godmode) " : "", message);
        write (fd, logstr, strlen (logstr));
        close (fd);
    }

    /* check high score file */

    if (GodMode) {
        I->p ("As god, you are ineligible for the highscore list.");
        return;
    }

    HighScoreEntry scores[NUMHIGHSCORES+1];
    int nscores = 0;
    int entry = 0;

    readHighScoreTable (scores, &nscores, &entry);

    if (NUMHIGHSCORES == entry) {
        /* didn't make a high score */
        scores[entry].mScore = mScore;
#ifdef _WIN32
        scores[entry].mUid = 0;
#else
        scores[entry].mUid = getuid ();
#endif
        strncpy (scores[entry].mName, mName, 30);
        strncpy (scores[entry].mMessage, message, 200);
    } else {
        if (++nscores > NUMHIGHSCORES) {
            nscores = NUMHIGHSCORES;
        }
        for (int i = nscores - 1; i > entry; --i) {
            memcpy (&scores[i], &scores[i-1], sizeof (HighScoreEntry));
        }
        scores[entry].mScore = mScore;
#ifndef _WIN32
        scores[entry].mUid = getuid ();
#else
        scores[entry].mUid = 0;
#endif
        strncpy (scores[entry].mName, mName, 30);
        strncpy (scores[entry].mMessage, message, 200);

        char filename[PRIME_PATH_LENGTH];
        snprintf (filename, sizeof(filename)-1, "%s/highscores.dat", SCOREDIR);
        fd = open (filename, O_WRONLY | O_TRUNC | O_CREAT | O_BINARY,
#ifdef _WIN32
                   S_IREAD | S_IWRITE);
#else
                   S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
#endif
        if (-1 == fd) {
            I->p ("Couldn't write high score file.");
            return;
        }
        for (int i = 0; i < nscores; ++i) {
            if (sizeof (HighScoreEntry) !=
                write (fd, &scores[i], sizeof (HighScoreEntry)))
            {
                I->p ("error writing high score file");
                close (fd);
                return;
            }
        }
        close (fd);
    }
}

#define SIZE 80
void
shHero::postMortem ()
{
    char mortemname[PRIME_PATH_LENGTH];
    int mortem;

    snprintf (mortemname, sizeof(mortemname)-1, "%s/mortem.txt", UserDir);

    mortem = open (mortemname, O_WRONLY | O_TRUNC | O_CREAT,
#ifdef _WIN32
                   S_IREAD | S_IWRITE);
#else
                   S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
#endif

    if (mortem == -1) {
        I->p ("Couldn't open mortem file.");
        I->pause ();
        // TODO: ask for alternate path and filename.
        return;
    }

    I->doScreenShot (mortem);

    if (Hero.mInventory->count ()) {
        write (mortem, "\n--- INVENTORY ---\n\n", 20);
    }
    shObjectType lastType = kUninitialized;
    shObject *obj;
    char buf[SIZE];
    int n = Hero.mInventory->count ();
    for (int i = 0; i < n; ++i) {
        obj = Hero.mInventory->get (i);
        if (obj->myIlk ()->mType != lastType) { /* Insert header. */
            lastType = obj->myIlk ()->mType;
            write (mortem, "  ", 2);
            write (mortem, objectTypeHeader[lastType],
                strlen (objectTypeHeader[lastType]));
            write (mortem, "\n", 1);
        }
        obj->identify ();
        snprintf (buf, SIZE, "%c - %s\n", obj->mLetter, obj->inv ());
        write (mortem, buf, strlen (buf));
    }

    int powers = 0;
    for (int i = kNoMutantPower; i < kMaxHeroPower; ++i) {
        if (mMutantPowers[i]) {
            if (powers == 0) {
                write (mortem, "\n--- MUTANT POWERS ---\n\n", 24);
                write (mortem, "  Power            Level Chance Status\n", 39);
                powers = 1;
            }
            snprintf (buf, SIZE, "%-18s  %2d    %3d%%   %s\n",
                MutantPowers[i].mName, MutantPowers[i].mLevel,
                getMutantPowerChance (i),
                mMutantPowers[i] > 1 ? "(on)" : "");
            write (mortem, buf, strlen (buf));
        }
    }

    int skills = 0;
    shSkillCode lastcode = kNoSkillCode;
    n = mSkills.count ();
    for (int i = 0; i < n; ++i) {
        shSkill *skill = mSkills.get (i);
        if (skill->mAccess == 0 || (skill->getValue () == 0)) continue;
        if (!skills) {
            write (mortem, "\n--- SKILLS ---\n\n", 17);
        }
        ++skills;
        if (lastcode == kNoSkillCode) {
            write (mortem, "Combat Skills\n", 14);
        } else if (lastcode & kWeaponSkill && !(skill->mCode & kWeaponSkill)) {
            write (mortem, "Adventuring Skills\n", 19);
        } else if (lastcode != kMutantPower && skill->mCode == kMutantPower) {
            write (mortem, "Mutant Power Skills\n", 20);
        }
        skill->getDesc (buf, SIZE);
        write (mortem, buf, strlen (buf));
        write (mortem, "\n", 1);
        lastcode = skill->mCode;
    }
    if (mSkillPoints) {
        sprintf (buf, "\nYou had %d unused skill point%s.\n",
            mSkillPoints, mSkillPoints > 1 ? "s" : "");
        write (mortem, buf, strlen (buf));
    }

    write (mortem, "\n--- DIAGNOSTICS ---\n\n", 22);
    Hero.doDiagnostics (0, 0, mortem);

    write (mortem, "\n--- KILL LIST ---\n\n", 20);
    for (int i = 0; i < kNumMonIlks; ++i) {
        shMonsterIlk *ilk = MonIlks[i];
        if (ilk->mKills) {
            if (ilk->mFeats & kUniqueMonster) {
                snprintf (buf, SIZE, "the %s\n", ilk->mName);
            } else {
                /* Need another buffer just for monster name because
                   number of slain creatures would confuse makePlural. */
                char monname[SIZE];
                snprintf (monname, SIZE, "%s", ilk->mName);
                if (ilk->mKills > 1) {
                    makePlural (monname, SIZE);
                }
                snprintf (buf, SIZE, "%d %s\n", ilk->mKills, monname);
            }
            write (mortem, buf, strlen (buf));
        }
    }

    write (mortem, "\n--- SUMMARY ---\n\n", 18);
    time_t clocktime;
    time (&clocktime);
    struct tm *when = localtime (&clocktime);

    snprintf (buf, SIZE, "%s finished adventuring on %d-%d-%d.\n",
        mName, when->tm_year + 1900, when->tm_mon + 1, when->tm_mday);
    write (mortem, buf, strlen (buf));
    snprintf (buf, SIZE,
        "%s earned %d experience points and achieved level %d.\n",
        mName, mXP, mCLevel);
    write (mortem, buf, strlen (buf));
    snprintf (buf, SIZE, "%s attained the professional rank of %s.\n",
        mName, getTitle ());
    write (mortem, buf, strlen (buf));
    if (GodMode) {
        snprintf (buf, SIZE, "Games in god mode are not scored. %d score points go down the drain.\n", mScore);
    } else {
        snprintf (buf, SIZE, "%s collected %d score points.\n",
            mName, mScore);
    }
    write (mortem, buf, strlen (buf));
    close (mortem);
}
#undef SIZE
